# -*- coding: utf-8 -*-
"""`bottle_jwt.auth` module.

Main auth providers class implementation.
"""

from __future__ import unicode_literals
from __future__ import print_function

import abc
import json
import six

__all__ = ['BaseBackend', 'FileSystemBackend']


class BackendError(Exception):
    """Raises when a backend error occurs.
    """


@six.add_metaclass(abc.ABCMeta)
class BaseBackend(object):
    """Auth Provider Interface.
    """

    @abc.abstractmethod
    def authenticate_user(self, user_uid, user_secret):  # pragma: no cover
        """User authentication method. All subclasses must implement the
        `authenticate_user` method with the following specs.

        Args:
            user_uid (str): User identity for the backend (email/username).
            user_secret: User secret (password) for backend.

        Returns:
            User id if authentication is succesful or None
        """
        pass

    @abc.abstractmethod
    def get_user(self, user_uid):  # pragma: no cover
        """User data retrieval method. All subclasses must implement the
        `get_user` method with the following specs.

        Args:
            user_uid (str): User identity in backend.

        Returns:
            User data (dict) if user exists or None.
        """
        pass


class FileSystemBackend(BaseBackend):
    """A JSON file-system backend system (useful for development only).
    Reads a json file that conforms to the following api:

    {"users": [
        {
            "user_id": "<user_id>",
            "password": "<password>",
            "attr1": value1,
            ...
            ...
        }
    ]
    }

    """

    def __init__(self, storage=''):
        self.storage = storage
        self._data = None

    @property
    def data(self):
        if not self._data:
            try:
                with open(self.storage, 'r') as db_storage:
                    data = db_storage.read()
                self._data = json.loads(data).get('users')

            except IOError as e:
                raise BackendError(e.args)
        return self._data

    def authenticate_user(self, user_uid, user_secret):
        """Implement `BaseBackend.authenticate_user` method.
        """

        try:
            return [user for user in self.data
                    if user['user_id'] == user_uid and
                    user['password'] == user_secret].pop()

        except IndexError:
            return None

    def get_user(self, user_uid):
        """Implement `BaseBackend.get_user` method.
        """
        user = [user for user in self.data if user['user_id'] == user_uid]
        if user:
            return user.pop()
        return None



# -*- coding: utf-8 -*-
"""Unit test fixtures for `bottle_jwt.backend` module
"""
from __future__ import unicode_literals
from __future__ import print_function


import pytest
import os
from bottle_jwt.backends import (FileSystemBackend, BackendError)


@pytest.fixture(scope='module')
def file_backend():
    """Fixture for `bottle_jwt.backends.FilesystemBackend` class.
    """

    pwd = '/'.join(os.path.abspath(__file__).split('/')[:-1])

    db_file = pwd + '/mock_db.json'

    return FileSystemBackend(db_file)


def test_filesystembackend_lazy_load_data_prop(file_backend):
    """Test `bottle_jwt.backends.FilesystemBackend` data property.
    """

    assert file_backend.data[0]['user_id'] == 'pav'


def test_filesystembackend_lazy_load_fail(file_backend):
    """Test `bottle_jwt.backends.FilesystemBackend` data property.
    """

    backend_cls = file_backend.__class__

    with pytest.raises(BackendError):
        assert backend_cls('/sds').data


def test_filesystembackend_authenticate_user_hit(file_backend):
    """Test `bottle_jwt.backends.FilesystemBackend.authenticate_user` method.
    """

    assert file_backend.authenticate_user('pav', '123')


def test_filesystembackend_get_user_not_found(file_backend):
    """Test `bottle_jwt.backends.FilesystemBackend.authenticate_user` method
    fail.
    """

    assert not file_backend.authenticate_user('pav', '12')



# -*- coding: utf-8 -*-
"""Unit test fixtures for `bottle_jwt.auth` module
"""
from __future__ import unicode_literals
from __future__ import print_function


import pytest
import bottle
from bottle_jwt.auth import (
    JWTProvider, JWTProviderError, JWTProviderPlugin, jwt_auth_required, JWTReplier
)


@pytest.fixture(scope='module')
def provider(backend):
    """Fixture for `bottle_jwt.JWTProvider` class instance.
    """
    b = backend({"pav": "123", "ama": "123", "max": '456'})

    return JWTProvider(
        fields=('username', 'secret'),
        backend=b,
        secret='123',
        ttl=1
    )


@pytest.fixture(scope='module')
def provider_plugin(backend):
    """Fixture for `bottle_jwt.JWTProviderPlugin` class instance.
    """

    b = backend({"pav": "123", "ama": "123", "max": '456'})

    return JWTProviderPlugin(
        'auth',
        auth_endpoint='/auth',
        fields=('username', 'secret'),
        backend=b,
        secret='123',
        ttl=1
    )


@pytest.fixture(scope='module')
def provider_plugin_replier():
    """Fixture for `bottle_jwt.JWTReplier` class instance.
    """

    return JWTReplier()


def test_provider_authenticate(provider, request):
    """
    """

    r = request({"username": "pav", "secret": "123"})

    token = provider.authenticate(r)

    assert len(token) > 100


def test_provider_authenticate_fail(provider, request):
    """
    """

    r = request({"username": "pax", "secret": "123"})

    with pytest.raises(JWTProviderError):
        provider.authenticate(r)


def test_provider_authorize_pass(provider, request):
    """
    """
    r = request({"username": "pav", "secret": "123"})

    token = provider.authenticate(r)

    r.set_header('Authorization', 'JWT {}'.format(token.decode("utf-8")))

    data = provider.authorize(r)

    assert data


def test_provider_authorize_invalid_token_fail(provider, request):
    """
    """
    r = request({"username": "pav", "secret": "123"})

    token = provider.authenticate(r)

    r.set_header('Authorization', 'JWT {}123'.format(token.decode("utf-8")))

    with pytest.raises(JWTProviderError):
        provider.authorize(r)


def test_provider_authorize_no_token_fail(provider, request):
    """
    """
    r = request({"username": "pav", "secret": "123"})

    with pytest.raises(JWTProviderError):
        provider.authorize(r)


def test_provider_authorize_exp_fail(provider, request):
    """
    """
    r = request({"username": "pav", "secret": "123"})

    token = provider.authenticate(r)

    r.set_header('Authorization', 'JWT {}'.format(token))

    import time

    time.sleep(2)

    with pytest.raises(JWTProviderError):
        provider.authorize(r)


def test_provider_plugin_register(provider_plugin):
    """
    """
    app = bottle.Bottle()

    app.install(provider_plugin)

    @jwt_auth_required
    def app_handler():
        return 'Private Resource'

    app.get('/')(app_handler)

    assert len(app.routes) == 2


def test_provider_plugin_replies(provider_plugin_replier):
    """
    This will test that the stock standard success reply is maintained and works
    """
    test_token = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
    result = provider_plugin_replier.auth_succeeded(test_token)
    assert (isinstance(result, dict))
    assert (result["token"] == test_token)
    """
    This will test that the stock standard failure reply is maintained and works
    """
    fake_errors = ["Test"]
    result = provider_plugin_replier.auth_failed(fake_errors)
    assert (isinstance(result, dict))
    assert (result["AuthenticationError"] == fake_errors)
    """
    Since the default 401 reply is an abort, let's make sure we catch it
    """
    exception_code = None
    try:
        provider_plugin_replier.auth_required(fake_errors)
    except bottle.HTTPError as e:
        exception_code = e.status_code
    assert(exception_code is not None and exception_code == 401)


from distutils.core import setup

setup(
    name='bottle-jwt',
    version='0.0.1',
    packages=['bottle_jwt'],
    url='',
    license='GLPv3',
    author='Papavassiliou Vassilis',
    author_email='vpapavasil@gmail.com',
    description='JWT Auth plugin for bottle.py applications'
)
